ExpStats = LibStub("AceAddon-3.0"):NewAddon("ExpStats", "AceConsole-3.0", "AceEvent-3.0","AceTimer-3.0")

local L = LibStub("AceLocale-3.0"):GetLocale("ExpStats", true)

local ex = ExpStats
local time = time

local LibQTip = LibStub("LibQTip-1.0")

local defaults = {
  profile = {
    dataValue = "XPHour",
  }
}

local xptypes = {
          XPHour = L["XP gain rate (session)"],
          XPHourLevel = L["XP gain rate (level)"],
          XP = L["Current XP"],
          XPpcnt = L["Percentage XP"],
          XPtotal = L["XP gained and total for level"],
          LastLevelGainTime = L["Last level gained"],
          PsuedoInstantaneousXPHour = L["Instantaneous XP rate"],
          XPsession = L["XP gained this session"],
          XPtolevel = L["XP until the next level"],
          TimeToLevel = L["The time at current rate until next level"],
          MobKillsToLevel = L["Mob kills at current rate until next level"],
          LastKill = L["Last kill XP and time"],
          RestExpiryETA = L["The time at current rate until rest runs out"],
          RestXP = L["Amount of rested XP"],
          RestXPpct = L["Percentage of rested XP"],
          BubblesOfLevel = L["Number of bubbles in level"],
          BubblesToLevel = L["Number of bubbles to next level"],
          BubbleRestToGo = L["Number of bubbles of rest"],
          RawMobKillsToLevel = L["Number of mob kills to level"],
}

local tooltipordering = {
          "XPtotal",
          "RestXP",
          "XPpcnt",
          "RestXPpct",
          "XPHour",
          "XPsession",
          "XPHourLevel",
          "LastLevelGainTime",
          "PsuedoInstantaneousXPHour",
          "XPtolevel",
          "TimeToLevel",
          "MobKillsToLevel",
          "RestExpiryETA",
          "LastKill",
          "RawMobKillsToLevel",
          "BubblesOfLevel",
          "BubblesToLevel",
          "BubbleRestToGo",
}

local options = {
    name = L["ExpStats"],
    handler = ex,
    type = "group",
  	get = function(info) return ex.db[info[#info]] end,
  	set = function(info, value) ex.db[info[#info]] = value end,
    args = {
      dataValue = {
        name = L["Main Value"],
        desc = L["Select the primary value to show"],
        type = "select",
        values = xptypes,
      },
    },
}

function ExpStats:OnInitialize()
  if UnitLevel("player")==GetMaxPlayerLevel() then
    self:Print(L["You are maximum level, disabling ExpStats"])
    self:SetEnabledState(false)
    return
  else
    self:SetEnabledState(true)
  end
  self.db_root = LibStub("AceDB-3.0"):New("ExpStatsDB",defaults)
  self.db=self.db_root.profile
	local acreg = LibStub("AceConfigRegistry-3.0")
	acreg:RegisterOptionsTable("ExpStats", options)
	acreg:RegisterOptionsTable("ExpStats Profiles", LibStub("AceDBOptions-3.0"):GetOptionsTable(self.db_root))
	local acdia = LibStub("AceConfigDialog-3.0")
	acdia:AddToBlizOptions("ExpStats", L["ExpStats"])
	acdia:AddToBlizOptions("ExpStats Profiles", L["Profiles"], "ExpStats")
	
	self.levelledXP = 0
	self.sessionXP = 0
	self.playedTimeOffset = time()
	self.lastLevelTimeOffset = time()
	self.sessionStartTime = time()
	
	self:buildLDBs()
end

local lastXP
local recentXPGains={}

function ExpStats:OnEnable()
	self:RegisterEvent("PLAYER_XP_UPDATE")
	self:RegisterEvent("PLAYER_LEVEL_UP")
	self:RegisterEvent("TIME_PLAYED_MSG")

  RequestTimePlayed()

	self:ScheduleRepeatingTimer("UpdateText",1)
	
	if not self.sessionStartXP then
		self.sessionStartXP = UnitXP("player")
	end
	lastXP = self.sessionStartXP
  table.insert(recentXPGains,{time=time(),xp=lastXP,xpgain=0})
  self.lastLevelMaxXP=UnitXPMax("player")
  self.tooltip = LibQTip:Acquire("ExpStatsTooltip", 2, "LEFT", "RIGHT")
  self:BuildTooltip()
end

function ExpStats:OnDisable()
	self:UnregisterEvent("PLAYER_XP_UPDATE")
	self:UnregisterEvent("PLAYER_LEVEL_UP")
	self:UnregisterEvent("TIME_PLAYED_MSG")
end

function ExpStats:PLAYER_XP_UPDATE(event, ...)
	local newXP = UnitXP("player")
  local lastMobXP = newXP - lastXP
	lastXP = newXP
	if lastMobXP < 0 then
		lastMobXP = 0
	end
	if not self.levellingFlag then
  	self.sessionXP = newXP - self.sessionStartXP + self.levelledXP
  end
	if lastMobXP > 0 then
	  table.insert(recentXPGains,{time=time(),xp=newXP,xpgain=lastMobXP,rest=GetXPExhaustion()})
	end
end

function ExpStats:PLAYER_LEVEL_UP(event, ...)
	self.lastLevelTimeOffset = time()
	self.levellingFlag=true
	self:ScheduleTimer("UpdateLevelInfo",1)
end

function ExpStats:UpdateLevelInfo(event, ...)
	for idx,hist in pairs(recentXPGains) do
	  hist.xp=hist.xp-self.lastLevelMaxXP
	end
  self.levelledXP = self.levelledXP + self.lastLevelMaxXP - self.sessionStartXP
  self.lastLevelMaxXP= UnitXPMax("player")
	self.sessionStartXP = 0
 	self.sessionXP = UnitXP("player") - self.sessionStartXP + self.levelledXP
	self.levellingFlag=false
end

function ExpStats:TIME_PLAYED_MSG(event, ...)
  local played, lastLevel = select(1,...)
	self.playedTimeOffset = time() - played
	self.lastLevelTimeOffset = time() - lastLevel
end

function ExpStats:UpdateText()
  local typ=self.db.dataValue or "XPHour"
  local fmt,unit,value,value2,value3=self['meth_'..typ](self)
  if unit then
    self.obj.value=value
    self.obj.unit=unit
  else
    self.obj.value=nil
    self.obj.unit=nil
  end
  self.obj.text=format(fmt,value,value2,value3)
  self:UpdateTooltip()
end

local highlightFont

function ExpStats:UpdateTooltip()
  if not self.showing then return end
  local curr=self.db.dataValue or "XPHour"
  local fnt
  for idx,k in ipairs(tooltipordering) do
    local fmt,_,value,value2,value3=self["meth_"..k](self)
    if k==curr then
      fnt=highlightFont
    else
      fnt=self.tooltip:GetFont()
    end
    self.tooltip:SetCell(idx+1,1,xptypes[k],fnt)
    self.tooltip:SetCell(idx+1,2,format(fmt,value,value2,value3),fnt)
  end
end

function ExpStats:BuildTooltip()
  self.tooltip:AddHeader(L["ExpStats"])
  for _,k in pairs(tooltipordering) do
    local label=xptypes[k]
    self.tooltip:AddLine(label,"N/A")
  end
  self.tooltip:AddLine(" ")
  self.tooltip:AddLine("Left click: next line","Right click: options")
  highlightFont = CreateFont("ExpStatsHighlightFont")
  highlightFont:CopyFontObject(self.tooltip:GetFont())
  highlightFont:SetTextColor(1,0,0)
end

-- XP computation methods

function ExpStats:meth_XP()
  return "%d XP", "XP", UnitXP("player")
end

function ExpStats:meth_XPpcnt()
  return "%.2f %%", "%", (UnitXP("player")/UnitXPMax("player")) *100
end

function ExpStats:meth_XPtotal()
  local xp,max=UnitXP("player"),UnitXPMax("player")
  return "%d/%d XP",nil,xp,max
end

function ExpStats:meth_XPHour()
  return "%.2f "..L["XP/hour"],L["XP/hour"],self.sessionXP / (time() - self.sessionStartTime) * 3600
end

function ExpStats:meth_XPHourLevel()
  return "%.2f "..L["XP/hour"],L["XP/hour"],UnitXP("player") / (time() - self.lastLevelTimeOffset) * 3600
end

function ExpStats:meth_LastLevelGainTime()
  return "%.2f "..L["hours ago"], L["hours ago"], (time() - self.lastLevelTimeOffset) / 3600
end

function ExpStats:meth_XPsession()
  return "%d XP","XP", self.sessionXP
end

function ExpStats:meth_XPtolevel()
  return "%d XP","XP", UnitXPMax("player")-UnitXP("player")
end

function ExpStats:meth_BubbleRestToGo()
  return "%.2f bubbles","bubbles", 20.0*(GetXPExhaustion() or 0)/UnitXPMax("player")
end

function ExpStats:meth_BubblesOfLevel()
  return "%.2f bubbles","bubbles", 20.0 * (UnitXP("player")/UnitXPMax("player"))
end

function ExpStats:meth_BubblesToLevel()
  return "%.2f bubbles","bubbles", 20.0 * (UnitXPMax("player")-UnitXP("player"))/UnitXPMax("player")
end

function ExpStats:meth_LastKill()
  local last=recentXPGains[#recentXPGains]
  return "%d XP %d " .. L["seconds ago"], nil, last.xpgain,time()-last.time
end

function ExpStats:meth_RawMobKillsToLevel()
  local last=recentXPGains[#recentXPGains]
  if last.xpgain>0 then
    return "%d kills","kills", (UnitXPMax("player")-UnitXP("player"))/last.xpgain
  else 
    return "Insufficient data",nil,0
  end
end

function ExpStats:meth_RestXP()
  return "%d XP","XP",GetXPExhaustion() or 0
end

function ExpStats:meth_RestXPpct()
  return "%.2f %%","%",100*(GetXPExhaustion() or 0)/UnitXPMax("player")
end

local function computeXP(history)
  local timespan=0.00000001
  local xpgain=0
  local now=time()
  local currXP=UnitXP("player")
  while #recentXPGains>1 and recentXPGains[1].time+history<now do
    table.remove(recentXPGains,1)
  end
  if #recentXPGains>1 then
    oldest=recentXPGains[1]
    timespan=now-oldest.time
    xpgain=currXP-oldest.xp
  end
  local currentRate=3600*(xpgain/timespan)
  if #recentXPGains>2 then
    local penultimate=recentXPGains[#recentXPGains-1]
    timespan=now-penultimate.time
    xpgain=currXP-penultimate.xp
  end
  local trendRate=3600*(xpgain/timespan)
  return currentRate,trendRate-currentRate,xpgain,#recentXPGains-1
end

function ExpStats:meth_TimeToLevel()
  local xprate,trend,_,cnt=computeXP(300)
  if cnt>0 then
    local trendstring="no trend"
    if trend>0 then
      trendstring="trending quicker"
    elseif trend<0 then
      trendstring="trending slower"
    end
    local deficit=UnitXPMax("player")-UnitXP("player")
    local hr,mn=math.modf(deficit/xprate)
    return "%dh %dm (%s)",nil,hr,math.floor(mn*60),trendstring
  else
    return L["Insufficient data"],nil,0
  end
end

function ExpStats:meth_MobKillsToLevel()
  local _,trend,xpgain,cnt=computeXP(300)
  if cnt>0 then
    local trendstring="no trend"
    if trend>0 then
      trendstring="trending less"
    elseif trend<0 then
      trendstring="trending more"
    end
    local deficit=UnitXPMax("player")-UnitXP("player")
    local mobKills=math.ceil((cnt*deficit)/xpgain)
    return "%d mobs (%s)","mobs",mobKills,trendstring
  else
    return L["Insufficient data"],nil,0
  end
end
  
function ExpStats:meth_PsuedoInstantaneousXPHour()
  local rate,trend=computeXP(300)
  local trendstring="0"
  if trend>0 then
    trendstring="+"
  elseif trend<0 then
    trendstring="-"
  end
  return "%.1f XP/hour (%s)","XP/hour",rate,trendstring
end

function ExpStats:meth_RestExpiryETA()
  local xprate,trend,_,cnt=computeXP(300)
  if cnt>0 then
    local trendstring="no trend"
    if trend>0 then
      trendstring="trending quicker"
    elseif trend<0 then
      trendstring="trending slower"
    end
    local deficit=GetXPExhaustion() or 0
    local hr,mn=math.modf(deficit/xprate)
    return "%dh %dm (%s)",nil,hr,math.floor(mn*60),trendstring
  else
    return L["Insufficient data"],nil,0
  end
end

function ExpStats:buildLDBs()
  local ldb = LibStub:GetLibrary("LibDataBroker-1.1", true)
  if ldb then 
	  ExpStats.obj=ldb:NewDataObject("ExpStats", {
		  type = "data source",
		  icon = "Interface\\AddOns\\expstats\\icon",
		  text = "",
		  OnClick = function(self, button)
		    if button=="RightButton" then
    			InterfaceOptionsFrame_OpenToCategory(L["ExpStats"])  
    	  elseif button=="LeftButton" then
    	    local curr=ExpStats.db.dataValue or "XPHour"
    	    local newidx=1
    	    for idx = 1, #tooltipordering-1 do
    	      if tooltipordering[idx]==curr then
    	        newidx=idx+1
    	        break
    	      end
    	    end
    	    ExpStats.db.dataValue=tooltipordering[newidx]
    		  ExpStats.showing=true
		      ExpStats:UpdateTooltip()
		      ExpStats.tooltip:Show()
        end
		  end,
		  OnEnter = function(self)
		    ExpStats.tooltip:SmartAnchorTo(self)
		    ExpStats.showing = true
		    ExpStats:UpdateTooltip()
		    ExpStats.tooltip:Show()
		  end,
		  OnLeave = function(self)
		    ExpStats.tooltip:Hide()
		    ExpStats.showing = false
		  end,
	  })
  end
end
